import os
import sys
import getopt
import stems
import bulges_internal
import pk_construction
import cc06
import cc09
import longPK
import pk_interrupted
import pk_energy
import mwis
import pk_tools

##################### Main Program #####################
local_pks = False
# -l: show best local pseudoknots
def Usage ():
    print "Usage"
    print "dotknot.py <infile> <outfile> [-l]"
    print "[-l] show best local pseudoknots"
    sys.exit(0)
#---------------------------------------------------------------------------------------------------------   
# Read fasta file from user input in command line
try:
    fasta_file = sys.argv[1]
    if sys.argv[1] != '-h':
        try:
            f = open(fasta_file,'U')
        except:
            print "Input file does not exist!"
            sys.exit(0)
    else:
        Usage()
except:
    print "Provide input file!"
    sys.exit(0)
	
# Read output file from user input in command line 
if len(sys.argv) > 2 and sys.argv[2] != '-l' and sys.argv[2] != '-h':
    output_file_name = sys.argv[2]
else:
    output_file_name = "pseudoknots.txt"

if len(sys.argv) > 2:
	if sys.argv[2] == '-l':
		local_pks = True
if len(sys.argv) > 3:
	if sys.argv[3] == '-l':
		local_pks = True
    	
f = open(fasta_file,'U')
identifier = f.readline()
seq = f.readline()
seq = seq.upper()
seq = seq.strip()
seq = seq.replace('T','U')

infile = open("input.fasta","w")
print >> infile, seq.strip()
infile.close()
print identifier.strip()
print seq

print "Sequence length: ", len(seq)
print
os.popen('ViennaRNA-1.8.4/Progs/RNAfold -p2 -d2 -noLP < input.fasta')
bp_stack_dic = stems.postscript(seq)
#---------------------------------------------------------------------------------------------------------            
# Construct stems from stack probabilities and store with mean probability in stem dictionary
matrix_stems = stems.find_stems(bp_stack_dic)
#---------------------------------------------------------------------------------------------------------
# Assign a local weight to the stems using enf energy evaluation
# Format is the following (start, end): (length, confidence, stack_energy, loop_energy)
matrix_stems = stems.evaluation(matrix_stems,seq)
cutoff_stack = 0.0 
cutoff_loop = 4.0 
matrix_stems = stems.filter_stems(matrix_stems, cutoff_stack, cutoff_loop)
#---------------------------------------------------------------------------------------------------------
# Construct stems with bulge loops, internal loops and multiloops 
cutoff_prob =  0.001
matrix_stems_ib = stems.filter_stems_prob(matrix_stems, cutoff_prob)
structures_dic = bulges_internal.internal_mwis(matrix_stems_ib)
bulge_internal_dic,multiloops = bulges_internal.evaluation_bulge_internal(structures_dic, seq)
cutoff_loops = 0.0
bulge_internal_dic = bulges_internal.filter_stems(bulge_internal_dic,cutoff_loops)
multiloops = bulges_internal.filter_stems(multiloops,cutoff_loops)
#---------------------------------------------------------------------------------------------------------
# Construct H-type core pseudoknots
stems_shortened,pseudoknot_dic = pk_construction.build_pseudoknots(matrix_stems)
# Construct pseudoknots with stems s_ib
stems_shortened_ib,pk_with_IB = pk_construction.pseudoknots_with_IB(bulge_internal_dic, matrix_stems)
#---------------------------------------------------------------------------------------------------------
# Re-evaluate stem energies to account for shortended stems
stems_shortened.update(stems_shortened_ib)
if stems_shortened:
    matrix_stems, stems_shortened_dic = stems.re_evaluation_dangling(stems_shortened,matrix_stems,seq)
else:
    stems_shortened_dic = {}
stems_shortened_dic = stems.filter_stems(stems_shortened_dic, cutoff_stack, cutoff_loop)
#---------------------------------------------------------------------------------------------------------
# Construct three different pseudoknot dictionaries
pseudoknot_dic_cc06,pseudoknot_dic_cc09,pseudoknot_dic_longpk = pk_construction.pk_dic_scan(pseudoknot_dic,matrix_stems,stems_shortened_dic)
#---------------------------------------------------------------------------------------------------------
# Heuristic energy parameters
init = 7.0 
penalty = 0.1
#---------------------------------------------------------------------------------------------------------
# Evaluate energies for pseudoknots with loop L2 <= 1nt
pseudoknot_dic_cc06_result = {}
pseudoknot_dic_cc06_result = cc06.dic_caochen06(pseudoknot_dic_cc06,matrix_stems,stems_shortened_dic,seq)
# Evaluate energies for pseudoknots with loop L2 >= 2nt and L2 <= 6nt
pseudoknot_dic_cc09_result = {}
pseudoknot_dic_cc09_result, entropies_dic, entropies_dic_L1, entropies_dic_L3 = cc09.dic_caochen09(pseudoknot_dic_cc09,matrix_stems)
# Evaluate energies for pseudoknots with loop L2 >= 7nt
pseudoknot_dic_longpk_result = {}
pseudoknot_dic_longpk_result = longPK.dic_longpks(pseudoknot_dic_longpk,matrix_stems,init,penalty)
# Now we form a pseudoknot dictionary out of the three results
pk_core_dic = {}
pk_core_dic.update(pseudoknot_dic_cc06_result)
pk_core_dic.update(pseudoknot_dic_cc09_result)
pk_core_dic.update(pseudoknot_dic_longpk_result)
#---------------------------------------------------------------------------------------------------------
# Now, look for recursive elements in the pseudoknot loops (MWIS) and add them to dictionary
pk_core_dic = pk_energy.recursive_pk(matrix_stems,bulge_internal_dic,multiloops,pk_core_dic)
# Combine recursive energies to overall pseudoknot energy
# Construct three different pseudoknot dictionaries for energy re-evaluation because of the loop entropies
pseudoknot_dic_cc06,pseudoknot_dic_cc09,pseudoknot_dic_longpk = pk_energy.pk_dic_scan_recursive(pk_core_dic,matrix_stems)  
pseudoknot_dic_cc06_result = cc06.pk_energy_reevaluation_06(pseudoknot_dic_cc06,matrix_stems)
pseudoknot_dic_cc09_result = cc09.pk_energy_reevaluation_09(pseudoknot_dic_cc09, matrix_stems, entropies_dic, entropies_dic_L1, entropies_dic_L3)
pseudoknot_dic_longpk_result = longPK.pk_energy_reevaluation_long(pseudoknot_dic_longpk,matrix_stems,init,penalty) 
pk_recursive_dic = {}
pk_recursive_dic.update(pseudoknot_dic_cc06_result)
pk_recursive_dic.update(pseudoknot_dic_cc09_result)
pk_recursive_dic.update(pseudoknot_dic_longpk_result)
#---------------------------------------------------------------------------------------------------------
# Evaluate energies for pseudoknots with stems s_ib
pk_dic_ib = pk_interrupted.evaluate_pk_with_IB(pk_with_IB, matrix_stems, stems_shortened_dic, init, penalty, seq)
pk_dic_ib = pk_interrupted.recursive_pk(matrix_stems,bulge_internal_dic,multiloops,pk_dic_ib)
pk_dic_ib = pk_interrupted.re_evaluate_pk_with_IB(pk_dic_ib, init, penalty)
pk_dic_ib = pk_energy.pk_filter(pk_dic_ib)
# Pseudoknots constructed with stems s_ib are stored in the pseudoknot dictionary
pk_recursive_dic.update(pk_dic_ib)
#---------------------------------------------------------------------------------------------------------
# Here, compute the best x pseudoknots with regards to two criteria
if local_pks:
    number_of_best_pks = 5
    best_pks_ratio = pk_tools.output_best_ratio(pk_recursive_dic, number_of_best_pks)
    best_pks_ratio = pk_tools.pk_structures(seq, best_pks_ratio, matrix_stems, stems_shortened_dic, bulge_internal_dic, multiloops, pk_recursive_dic, pk_dic_ib)
    best_pks_energy = pk_tools.output_best_energy(pk_recursive_dic, number_of_best_pks)
    best_pks_energy = pk_tools.pk_structures(seq, best_pks_energy, matrix_stems, stems_shortened_dic, bulge_internal_dic, multiloops, pk_recursive_dic, pk_dic_ib)
#---------------------------------------------------------------------------------------------------------
# Filtering step for all pseudoknots
pk_recursive_dic = pk_energy.pk_filter(pk_recursive_dic)
#---------------------------------------------------------------------------------------------------------
cutoff_prob =  0.001
matrix_stems_mwis = stems.filter_stems_prob(matrix_stems, cutoff_prob)

if matrix_stems_mwis or pk_recursive_dic or pk_dic_ib or bulge_internal_dic or multiloops:
    mwis_dic,detected_pks = mwis.method(matrix_stems_mwis, pk_recursive_dic, bulge_internal_dic, multiloops)
    pseudoknot_list = mwis.pk_structures(seq, detected_pks, matrix_stems, stems_shortened_dic, bulge_internal_dic, multiloops, pk_recursive_dic, pk_dic_ib)
else:
    pseudoknot_list = []

pk_file = open(output_file_name,"w")

if pseudoknot_list:
	print >> pk_file, "Detected pseudoknots:"
	for pk in pseudoknot_list:
		print >> pk_file, str(pk[0]),str(pk[1]), str(pk[2])
		print >> pk_file, str(pk[3])
		print >> pk_file, str(pk[4])		
if not pseudoknot_list:
    print >> pk_file, "No pseudoknots were detected."    
pk_file.close()

if local_pks:
	print "Show near-optimal pseudoknots"
	print
if local_pks and best_pks_ratio:
    print "Best", number_of_best_pks, "local pseudoknots in terms of energy to length ratio:"
    for item in best_pks_ratio:
        print item[0], item[1], item[2]
        print item[3]
        print item[4]
    print

if local_pks and best_pks_energy:
    print "Best", number_of_best_pks, "local pseudoknots in terms of free energy:"
    for item in best_pks_energy:
        print item[0], item[1], item[2]
        print item[3]
        print item[4]
    print

if local_pks and not best_pks_ratio and not best_pks_energy:
    print "No local pseudoknots were detected."
